
local server = {}

local joblist = require("sga.joblist")

local safer = require("safer")
local restserver = require("restserver")
local dkjson = require("dkjson")
local copas = require("copas")

local function create_action_uris(self, jid)
   local action_uris = {}
   local uri = self.base_uri.."/v1/sga/"..self.config.sga_name.."/job/"..jid
   for action, _ in pairs(self.driver.actions) do
      action_uris[action] = uri.."/"..action
   end
   return action_uris
end

local function cleanup_job(self, job)
   local pok, err = pcall(self.driver.cleanup_job, self.driver, job)
   if not pok then
      self.logger:error(err)
   end
   local ok = self.joblist:finish(job)
   if not ok then
      self.logger:error("Failed marking job "..job.jid.." as finished")
   end
end

local Server = safer.readonly {

   get_persisted_jobs = function(self)
      local persisted_jobs = { lost = {}, retrieved = {} }
      for job in self.joblist:each() do
         local is_done, err = self.driver:is_command_done(job)
         if is_done == nil then
            if not job.done then
               table.insert(persisted_jobs.lost, { cmd_id = job.cmd_id } )
               cleanup_job(self, job)
            end
         elseif not is_done then
            table.insert(persisted_jobs.retrieved, { cmd_id = job.cmd_id, actions = create_action_uris(self, job.jid) } )
         else
            cleanup_job(self, job)
         end
      end
      return persisted_jobs
   end,

   start = function(self)
      self.rs:enable("restserver.xavante"):start()
   end,

}

function server.new(config, logger, client, driver)

   local self = {
      rs = restserver:new():host(config.sgad_host):port(config.sgad_port),
      joblist = joblist.new(config),
      base_uri = "http://"..config.sgad_host..":"..config.sgad_port,
      config = config,
      logger = logger,
      client = client,
      driver = driver,
      down = false,
   }
   self = safer.table(setmetatable(self, { __index = Server }))

   local function launch_exec_polling_coro(job)
      copas.addthread(function()
         while (not job.done) and (not self.down) do
            copas.sleep(self.config.exec_polling_interval_s)
            local pok, is_done, walltime_s, usertime_s, systime_s = pcall(self.driver.is_command_done, self.driver, job)
            if not pok then
               self.logger:error(is_done)
               break
            elseif is_done then
               if walltime_s then
                  local ok = self.client:completion(job.cmd_id, walltime_s, usertime_s, systime_s) 
                  if ok then
                     cleanup_job(self, job)
                  end
                  self.logger:info("Command "..job.jid.." done")
               else
                  cleanup_job(self, job)
               end
               break
            elseif is_done == nil then
               self.logger:error(walltime_s)
               cleanup_job(self, job)
               break
            end
         end
      end)
   end

   for job in self.joblist:each() do
      if not job.done then
         launch_exec_polling_coro(job)
      end
   end

   local function succeed(msg, code)
      return restserver.response():status(code or 200):entity(msg)
   end

   local function fail(msg, code)
      self.logger:error(msg)
      return restserver.response():status(code or 400):entity(msg)
   end
   
   local function fail_sga_name(sga_name)
      return fail("Unknown sga "..sga_name, 404)
   end

   local function handle_action(sga_name, jid, action, ...)
      jid = tonumber(jid)
      self.logger:info("Received action '"..action.."' for job "..jid)
      if sga_name ~= config.sga_name then
         return fail_sga_name(sga_name)
      end
      local action_fn = self.driver.actions[action]
      if not action_fn then
         return fail("Action not supported", 405)
      end
      local job = self.joblist:get(jid)
      if not job then
         return fail("Job not found", 404)
      end
      local pok, ok, response, code = pcall(action_fn, self.driver, job, action, ...)
      if not pok then
         return fail(ok, 500)
      end
      if ok then
         return succeed(response)
      else
         return fail(response, code)
      end
   end

   self.rs:add_resource("v1/sga", {

      {
         method = "GET",
         path = "/{sga_name:[A-Za-z0-9_%-]+}/path",
         produces = "application/json",
         handler = function(query, sga_name)
            if sga_name ~= config.sga_name then
               return fail_sga_name(sga_name)
            end
            if not query.name then
               return fail("Missing 'name' argument.", 400)
            end
            self.logger:info("Received path query ("..query.name.." ).")
            -- Catch invalid characters
            local clean = query.name:match("^/[a-zA-Z0-9.@%-_ \128-\255/]*$")
            -- Catch attempts to escape using '..'
            if query.name:match("/%.%.") then
               clean = false
            end
            if not clean then
               return fail("Invalid characters in request.", 403)
            end
            -- Only allow peeking inside project and algorithm dirs
            if not (query.name:find(config.project_root_dir, 1, true) == 1 or query.name:find(config.algorithm_root_dir, 1, true) == 1) then
               return fail("Path unknown to SGA.", 404)
            end
            
            local pok, reply = pcall(self.driver.check_path, self.driver, query.name)
            if not pok then
               fail(reply, 500)
            elseif not reply then
               fail("Path not found ("..query.name..").", 404)
            end
            return succeed(reply)
         end,
      },
  
      {
         method = "POST",
         path = "/{sga_name:[A-Za-z0-9_%-]+}/job",
         consumes = "application/json",
         produces = "application/json",
         input_schema = {
            cmd_string = { type = "string" },
            cmd_id = { type = "string" },
            parameters = { type = "table" },
         },
         handler = function(job_request, sga_name)
            if sga_name ~= config.sga_name then
               return fail_sga_name(sga_name)
            end
            self.logger:info("Received job: "..job_request.cmd_string.." "..job_request.cmd_id.." "..dkjson.encode(job_request.parameters))

            local job = self.joblist:new_job(job_request.cmd_id, job_request.cmd_string, job_request.parameters)
            local pok, ok, err = pcall(self.driver.execute_command, self.driver, job, job_request.cmd_string)
            self.joblist:accept_job(job, pok and ok)
            if not pok then
               fail(ok, 500)
            elseif not ok then
               fail(err, 503)
            end
            self.joblist:save()
            launch_exec_polling_coro(job)
            local action_uris = create_action_uris(self, job.jid)
            local job_response = {
               cmd_id = job.cmd_id,
               actions = action_uris,
            }
            return succeed(job_response, 201)
         end,
      },
   
      {
         method = "GET",
         path = "/{sga_name:[A-Za-z0-9_%-]+}/job/{jid:[0-9]+}/{action:[a-z_][a-z0-9_]+}",
         produces = "application/json",
         handler = function(query, sga_name, jid, action)
            return handle_action(sga_name, jid, action, query)
         end,
      },

      {
         method = "POST",
         path = "/{sga_name:[A-Za-z0-9_%-]+}/job/{jid:[0-9]+}/{action:[a-z_][a-z0-9_]+}",
         consumes = "application/json",
         produces = "application/json",
         handler = function(post_data, sga_name, jid, action)
            return handle_action(sga_name, jid, action, post_data)
         end,
      },

      {
         method = "GET",
         path = "/{sga_name:[A-Za-z0-9_%-]+}/shutdown",
         produces = "application/json",
         handler = function(_, sga_name)
            if sga_name ~= config.sga_name then
               return fail_sga_name(sga_name)
            end
            self.logger:info("Received shutdown command...")
            -- FIXME Should we terminate all jobs?
            self.joblist:terminate()
            self.down = true
            copas.addthread(function()
               copas.sleep(2)
               self.logger:info("Shutting down Xavante...")
               self.rs:shutdown()
            end)
            return succeed({}, 200)
         end,
      },
      
   })

   return self
end

return server
